//
// Created by philipp on 01.12.17.
//

#ifndef PROTOCOL_TCPNETWORKSERVICECLIENT_H
#define PROTOCOL_TCPNETWORKSERVICECLIENT_H

#include <string>
#include <functional>
#include <boost/asio/io_service.hpp>
#include <boost/asio/ip/tcp.hpp>
#include <boost/asio/connect.hpp>
#include "Message.h"
#include "Networking.h"
#include "Utils.h"
#include "Error.h"

namespace networking
{
namespace service
{

template<typename Service>
class Client : public std::enable_shared_from_this<Client<Service>>
{
private:
    struct PrivateTag
    {
    };

public:
    using Ptr = std::shared_ptr<Client<Service>>;

    using RequestMessage = typename Service::RequestMessage;
    using ResponseMessage = typename Service::ResponseMessage;

    using Socket = typename boost::asio::ip::tcp::socket;

    using CallHandler = std::function<void(const error::ErrorCode & error, const ResponseMessage & response)>;

    static Ptr create(Networking & net)
    {
        return std::make_shared<Client<Service>>(PrivateTag{}, net);
    }

    Client(PrivateTag, Networking & net)
        : net(net)
          , socket(net.getIoService())
    {}

    ResponseMessage call(const std::string & host,
                         std::uint16_t port,
                         const RequestMessage & request,
                         time::Duration timeout)
    {
        if (calling)
            throw error::Busy{};

        calling = true;
        // Reset the calling flag when we leave this function using RAII.
        auto callingFlagReset = utils::RAIIObject{[this]
                                                  { calling = false; }};
        // Close the socket on leaving.
        closeable::Closer<Socket> socketCloser{socket};

        // We have three places to lose time: connecting, sending and receiving.
        // So after each operation we subtract the time spend from our timeout.
        auto startTime = time::now();
        newSocket();
        // Connect to server.
        networking::socket::connect(net, socket, host, port, timeout);
        updateTimeout(timeout, startTime);
        // Send the request.
        message::send(net, socket, request, timeout);
        updateTimeout(timeout, startTime);
        // Receive the response.
        return message::receive<ResponseMessage>(net, socket, timeout);
    }

    void asyncCall(const std::string & host,
                   std::uint16_t port,
                   const RequestMessage & request,
                   const time::Duration & timeout,
                   const CallHandler & handler)
    {
        if (calling)
            throw error::Busy{};

        try
        {
            newSocket();
            auto self = this->shared_from_this();

            // Container for our variables which are needed for the subsequent asynchronous calls to connect, receive and send.
            // When 'state' goes out of scope, it does cleanup and calls the handler and therefore makes use of the RAII concept.
            auto state = std::make_shared<AsyncState>(
                self, handler, false, ResponseMessage{ResponseMessage::emptyBuffer()}, error::codes::UNSPECIFIED,
                timeout, time::now());

            calling = true;

            // Connect to server.
            networking::socket::asyncConnect(
                net, socket, host, port, state->timeout,
                [self, state, request](const auto & error, auto && ... rest)
                {
                    // Only execute the handler if we're running on the networking thread.
                    state->shouldCallHandler = true;

                    if (error)
                    {
                        state->error = error;
                        return;
                    }

                    Client<Service>::updateTimeout(state->timeout, state->startTime);

                    // Send the request.
                    networking::message::asyncSend(
                        self->net, self->socket, request, state->timeout,
                        [self, state](const auto & error)
                        {
                            if (error)
                            {
                                state->error = error;
                                return;
                            }

                            Client<Service>::updateTimeout(state->timeout, state->startTime);

                            // Receive the response.
                            networking::message::asyncReceive<ResponseMessage>(
                                self->net, self->socket, state->timeout,
                                [self, state](auto const & error, const auto & response)
                                {
                                    state->error = error;
                                    state->responseMessage = response;
                                });
                        });
                });
        }
        catch (...)
        {
            throw error::FailedOperation{};
        }
    }

    bool isCalling() const noexcept
    {
        return calling;
    }

    void stop()
    {
        if (!calling)
            return;

        closeable::Closer<Socket>::close(socket);
    }

private:
    networking::Networking & net;
    Socket socket;
    std::atomic<bool> calling{false};

    static void updateTimeout(time::Duration & timeout, time::TimePoint & startTime)
    {
        auto nowTime = time::now();
        auto timeSpend = nowTime - startTime;
        startTime = nowTime;
        timeout -= timeSpend;
    }

    void newSocket()
    {
        if (!socket.is_open())
            socket = Socket(net.getIoService());
    }

    // We must keep track of some variables during the async handler chain.
    struct AsyncState
    {
        AsyncState(Ptr self,
                   CallHandler handler,
                   bool shouldCallHandler,
                   ResponseMessage responseMessage,
                   error::ErrorCode error,
                   time::Duration timeout,
                   time::TimePoint startTime)
            : self(self)
              , handler(handler)
              , shouldCallHandler(shouldCallHandler)
              , responseMessage(responseMessage)
              , error(error)
              , timeout(timeout)
              , startTime(startTime)
        {}

        Ptr self;
        CallHandler handler;
        bool shouldCallHandler;
        ResponseMessage responseMessage;
        error::ErrorCode error;
        time::Duration timeout;
        time::TimePoint startTime;

        ~AsyncState()
        {
            closeable::Closer<Socket>::close(self->socket);
            self->calling = false;
            if (shouldCallHandler)
            {
                try
                { handler(error, responseMessage); }
                catch (...)
                {}
            }
        }
    };
};

}
}


#endif //PROTOCOL_TCPNETWORKSERVICECLIENT_H
