//
// Created by philipp on 30.11.17.
//

#include <boost/asio/ip/tcp.hpp>
#include <iostream>
#include "../include/FollowerVehicle.h"
#include "../include/Utils.h"

using namespace protocol;
using namespace std::chrono_literals;

constexpr networking::time::Duration FollowerVehicle::HEARTBEAT_INTERVAL;
constexpr networking::time::Duration FollowerVehicle::BROADCAST_TIMEOUT;

FollowerVehicle::Ptr FollowerVehicle::create(networking::Networking & net, VehicleId vehicleId)
{
    return std::make_shared<FollowerVehicle>(PrivateTag{}, net, vehicleId);
}

FollowerVehicle::FollowerVehicle(PrivateTag, networking::Networking & net, VehicleId vehicleId)
    : Vehicle(net, vehicleId, Role::FOLLOWER)
{
    heartbeatTimer = networking::time::Timer::create(net);
    heartbeatSender = networking::message::DatagramSender<PlatoonMessage>::create(net);
    broadcastReceiver = networking::message::DatagramReceiver<PlatoonMessage>::create(net,
                                                                                      getOwnEndpoint().getUdpPort());
}

void FollowerVehicle::stop()
{
    Vehicle::stop();

    stopHeartbeat();

    stopReceiveBroadcast();
}

PlatoonConfig FollowerVehicle::getPlatoonConfig() const noexcept
{
    return platoonConfig.load();
}

void FollowerVehicle::doCreatePlatoon()
{
    if (state != State::IDLE)
        return;

    setState(State::CREATING_PLATOON);

    Vehicle::doCreatePlatoon();
}

void FollowerVehicle::doRunPlatoon()
{
    if (state != State::CREATING_PLATOON)
        return;

    setState(State::RUNNING_PLATOON);

    platoonCreateServiceServer->stop();

    startHeartbeat();

    startReceiveBroadcast();

    onRunningPlatoonCallback();
}

void FollowerVehicle::doLeavePlatoon()
{
    if (state == State::IDLE)
        return;

    stop();

    sendLeavePlatoonMessage();

    setState(State::IDLE);

    onLeavingPlatoonCallback();
}

void FollowerVehicle::makePlatoonCreateRequest(const VehicleEndpoint & vehicleEndpoint)
{
    // We'll send a request indicating that we want to be the follower.
    auto self = shared_from_this();
    platoonCreateServiceClient->asyncCall(
        vehicleEndpoint.getHost(),
        vehicleEndpoint.getTcpPort(),
        PlatoonMessage::followerRequest(getOwnEndpoint().getVehicleId()),
        3s,
        [self, vehicleEndpoint](const auto & error, const auto & response)
        {
            if (self->state != State::CREATING_PLATOON)
                return;

            if (error)
                return;

            if (response.getVehicleId() != vehicleEndpoint.getVehicleId() ||
                response.getMessageType() != messageTypes::ACCEPT_RESPONSE)
                return;

            self->leader = vehicleEndpoint;
            self->platoonId = response.getPlatoonId();
            self->doRunPlatoon();
        });
}

PlatoonMessage FollowerVehicle::makePlatoonCreateResponse(const tcp::endpoint & clientEndpoint,
                                                          const PlatoonMessage & request)
{
    if (state != State::CREATING_PLATOON)
        return PlatoonMessage::rejectResponse(getOwnEndpoint().getVehicleId());

    // Only consider this message if it is a leader request,
    // meaning that a leader wants us to join its platoon.
    if (request.getMessageType() != messageTypes::LEADER_REQUEST)
        return PlatoonMessage::rejectResponse(getOwnEndpoint().getVehicleId());

    auto leaderId = request.getVehicleId();
    // TODO: change vehicle endpoint information!
    leader = VehicleEndpoint{
        leaderId,
        clientEndpoint.address().to_string(),
        leaderId + 10000,
        leaderId + 10010
    };
    platoonId = request.getPlatoonId();

    doRunPlatoon();

    return PlatoonMessage::acceptResponse(getOwnEndpoint().getVehicleId(), platoonId);
}

void FollowerVehicle::startHeartbeat()
{
    auto self = shared_from_this();
    heartbeatTimer->startPeriodicTimeout(
        HEARTBEAT_INTERVAL,
        [self]
        {
            self->heartbeatSender->asyncSend(
                PlatoonMessage::heartbeatMessage(self->getOwnEndpoint().getVehicleId(), self->platoonId),
                self->leader.getHost(),
                self->leader.getUdpPort(),
                HEARTBEAT_INTERVAL,
                [self](auto ...)
                {});
        });
}

void FollowerVehicle::stopHeartbeat()
{
    heartbeatTimer->stop();
    heartbeatSender->stop();
}

void FollowerVehicle::startReceiveBroadcast()
{
    lastBroadcastTimestamp = networking::time::now();
    receiveBroadcast();
}

void FollowerVehicle::receiveBroadcast()
{
    auto self = shared_from_this();
    broadcastReceiver->asyncReceive(
        BROADCAST_TIMEOUT,
        [self](const auto & error, const PlatoonMessage & message, const auto & host, auto port)
        {
            if (self->state != State::RUNNING_PLATOON)
                return;

            auto timeSinceLastBroadcast = networking::time::now() - self->lastBroadcastTimestamp;

            // Broadcast-Timeout or manually stopped.
            if (error == networking::error::codes::ABORTED ||
                timeSinceLastBroadcast >= BROADCAST_TIMEOUT)
            {
                self->doLeavePlatoon();
                return;
            }
            else if (error ||
                     message.getVehicleId() != self->leader.getVehicleId() ||
                     message.getPlatoonId() != self->platoonId)
            {
                // In case of any other error we just ignore the message and receive again.
                self->receiveBroadcast();
                return;
            }

            if (message.getMessageType() == messageTypes::BROADCAST)
            {
                PlatoonConfig newConfig{message.getPlatoonSpeed(), message.getInnerPlatoonDistance()};
                if (self->platoonConfig.load() != newConfig)
                {
                    self->platoonConfig = newConfig;
                    self->onPlatoonConfigUpdatedCallback();
                }
                // Clear the broadcast-timeout.
                self->lastBroadcastTimestamp = networking::time::now();
            }

            // Receive the next broadcast.
            self->receiveBroadcast();
        });
}

void FollowerVehicle::stopReceiveBroadcast()
{
    broadcastReceiver->stop();
}

void FollowerVehicle::sendLeavePlatoonMessage()
{
    auto self = shared_from_this();
    heartbeatSender->asyncSend(
        PlatoonMessage::leaveMessage(getOwnEndpoint().getVehicleId(), platoonId),
        leader.getHost(),
        leader.getUdpPort(),
        HEARTBEAT_INTERVAL,
        [self](auto ...)
        {});
}

FollowerVehicle::Ptr FollowerVehicle::shared_from_this()
{
    return std::static_pointer_cast<FollowerVehicle>(Vehicle::shared_from_this());
}
